-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local beamstate = require('beamstate') --useful file that simplifies breaking stuff
local fire = require('fire') --used to ignite random nodes and explode vehicle

local damageTimerEnabled = false
local damageTimer = 0
local canDeflateTires --no default value, overwriting this variable is dangerous!
local hasNoTires
local timeBetweenFailures

local ignoreEngineMounts
local ignoreRadiatorAndOilpan
local toggleRandomExplosion
local toggleRealisticDamage
local engineDeformationOnly
local lowerTireDeflationChance

local breakRandomPartIgnoreWheels
local breakRandomBeamIgnoreWheels
local deformRandomBeamIgnoreWheels

--table with common prefixes of node names used for wheel centers
local genericWheelCenterNodeNames = 
{
	"fw",
	"rw",
	"stw",
	"fh",
	"rh",
	"bogw",
	"bt", --BELAZ mod
	"ax" --brute fix for Bluebuck, it uses 2 separate break groups for the wheels...
}

--table with common names or prefixes for powertrain part break groups, in vanilla cars and some mods
local genericPowertrainBreakGroupNames =
{
	"eng",
	"Eng",
	"transmission",
	"Transmission",
	"shaft",
	"Shaft",
	"fuel",
	"Fuel",
	"susp",
	"leaf",
	"Leaf",
	"subframe",
	"strut",
	"steering",
	"Steering",
	"arm",
	"Arm",
	"airbox",
	"shackle",
	"hub",
	"axle",
	"Axle",
	"damp"
}

--table with common engine deform groups
local genericEngineDamageDeformGroups = 
{
	"Engine",
	"oilpan",
	"radiator",
	"intake",
	"turbo",
	"intercooler",
	"piping",
	"accessories",
	"supercharger",
	"airbox"
}

--brittle parts that don't damage anything
local brittleGroupsToIgnore = 
{
	"radiator_damage",
	"oilpan_damage",
	"mainEngine",
	"shaft" --fits all driveshafts, intershafts, halfshafts
}


local M = {}


local function setSettings(setting)

  if setting == "ignoreEngineMounts" then ignoreEngineMounts = true end
  if setting == "ignoreRadiatorAndOilpan" then ignoreRadiatorAndOilpan = true end
  if setting == "toggleRandomExplosion" then toggleRandomExplosion = true end
  if setting == "toggleRealisticDamage" then toggleRealisticDamage = true end
  if setting == "engineDeformationOnly" then engineDeformationOnly = true end
  if setting == "lowerTireDeflationChance" then lowerTireDeflationChance = true end

end


local function unsetSettings(setting)

  if setting == "ignoreEngineMounts" then ignoreEngineMounts = false end
  if setting == "ignoreRadiatorAndOilpan" then ignoreRadiatorAndOilpan = false end
  if setting == "toggleRandomExplosion" then toggleRandomExplosion = false end
  if setting == "toggleRealisticDamage" then toggleRealisticDamage = false end
  if setting == "engineDeformationOnly" then engineDeformationOnly = false end
  if setting == "lowerTireDeflationChance" then lowerTireDeflationChance = false end

end


local function setWheelSettings(setting)

  if setting == "breakRandomPart" then breakRandomPartIgnoreWheels = true end
  if setting == "breakRandomBeam" then breakRandomBeamIgnoreWheels = true end
  if setting == "deformRandomBeam" then deformRandomBeamIgnoreWheels = true end

end


local function unsetWheelSettings(setting)

  if setting == "breakRandomPart" then breakRandomPartIgnoreWheels = false end
  if setting == "breakRandomBeam" then breakRandomBeamIgnoreWheels = false end
  if setting == "deformRandomBeam" then deformRandomBeamIgnoreWheels = false end

end


local function breakRandomPart() --breaks a random breakGroup

  local beams_with_breakGroups = {} --this will store all beams that can still trigger breakGroups
  
  if ignoreEngineMounts == false then --standard variant
  
    if breakRandomPartIgnoreWheels == false then --don't ignore wheels
	
	  for _, b in pairs(v.data.beams) do  --cycle through all beams
		 --check if a beam is a breakGroup trigger, belongs to exactly 1 breakGroup and isn't already broken
		if b.breakGroup ~= nil and type(b.breakGroup) ~= "table" and b.breakGroupType == 0 and not obj:beamIsBroken(b.cid) then
		  table.insert(beams_with_breakGroups, b) --then add it to the table
	    end
	 end
	   
	 else --ignore wheels
	 
	   for _, b in pairs(v.data.beams) do  --cycle through all beams
		 --check if a beam is a breakGroup trigger, belongs to exactly 1 breakGroup and isn't already broken
		 if b.breakGroup ~= nil and type(b.breakGroup) ~= "table" and b.breakGroupType == 0 and not obj:beamIsBroken(b.cid) then
		   --check if the beam is attached to what is a generic name for a wheel node, and if the breakGroup doesn't trigger wheel breakage otherwise
		   for _, n in pairs(genericWheelCenterNodeNames) do
		     if v.data.nodes[b.id1].name ~= nil and v.data.nodes[b.id2].name ~= nil and not v.data.nodes[b.id1].name:match(n) and not v.data.nodes[b.id2].name:match(n) and not string.match(b.breakGroup, "wheel") and not string.match(b.breakGroup, "Wheel") and not string.match(b.breakGroup, "axle") then	
			   table.insert(beams_with_breakGroups, b) --then add it to the table		
			 end
		   end
	     end
	   end
	   
     end
	 
   else --variant with setting enabled
   
     if breakRandomPartIgnoreWheels == false then --don't ignore wheels
	 
	  for _, b in pairs(v.data.beams) do  --cycle through all beams
		 --check if a beam is a breakGroup trigger, belongs to exactly 1 breakGroup and isn't already broken
		if b.breakGroup ~= nil and type(b.breakGroup) ~= "table" and b.breakGroupType == 0 and not obj:beamIsBroken(b.cid) then
		  --check all possible hints that could mean that the beam belongs to engine mounts or other powertrain parts
		  local beamBelongsToPowertrain = false
		  for _, g in pairs(genericPowertrainBreakGroupNames) do
		    if string.match(b.breakGroup, g) then
			  beamBelongsToPowertrain = true
			end
		  end
		  if beamBelongsToPowertrain == false then
			table.insert(beams_with_breakGroups, b) --then add it to the table
		  end
		end
	 end
	   
	 else --ignore wheels
	 
	   for _, b in pairs(v.data.beams) do  --cycle through all beams
		 --check if a beam is a breakGroup trigger, belongs to exactly 1 breakGroup and isn't already broken
		 if b.breakGroup ~= nil and type(b.breakGroup) ~= "table" and b.breakGroupType == 0 and not obj:beamIsBroken(b.cid) then
		  --check all possible hints that could mean that the beam belongs to engine mounts or other powertrain parts
		  local beamBelongsToPowertrain = false
		  for _, g in pairs(genericPowertrainBreakGroupNames) do
		    if string.match(b.breakGroup, g) then
			  beamBelongsToPowertrain = true
			end
		  end
		  if beamBelongsToPowertrain == false then
			 --check if the beam is attached to what is a generic name for a wheel node, and if the breakGroup doesn't trigger wheel breakage otherwise
			 for _, n in pairs(genericWheelCenterNodeNames) do
			   if v.data.nodes[b.id1].name ~= nil and v.data.nodes[b.id2].name ~= nil and not v.data.nodes[b.id1].name:match(n) and not v.data.nodes[b.id2].name:match(n) and not string.match(b.breakGroup, "wheel") and not string.match(b.breakGroup, "Wheel") and not string.match(b.breakGroup, "axle") then
				 table.insert(beams_with_breakGroups, b) --then add it to the table
			   end
		     end
		  end
	    end
	 end
	   
	end
	 
   end
  
  local random_beam = beams_with_breakGroups[math.random(#beams_with_breakGroups)] --select random beam to be broken
  if random_beam == nil then --if there is nothing else to break, don't break anything
    guihooks.message("Nothing left to break off that fits the current settings...", 10, "agenty_mechanical_failure_mod", "") --show message
    return 
  end
  
  local random_breakGroup = random_beam.breakGroup or "unknown" --get the selected beam's breakGroup
  obj:breakBeam(random_beam.cid) --break it
  guihooks.message(random_breakGroup .. " broken", 10, "agenty_mechanical_failure_mod", "") --show message what broke
  
end


local function deformRandomPart() --triggers a random deformGroup

  local beams_with_deformGroups = {} --this will store all beams that can still trigger deformGroups
  
  if ignoreRadiatorAndOilpan == false then
    for _, b in pairs(v.data.beams) do  --cycle through all beams
       --check if a beam is a deformGroup trigger, triggers exactly 1 deformGroup, and isn't broken
      if b.deformGroup ~= nil and type(b.deformGroup) ~= "table" and not obj:beamIsBroken(b.cid) then
	    table.insert(beams_with_deformGroups, b) --then add it to the table
      end
    end 
  else  
    for _, b in pairs(v.data.beams) do  --cycle through all beams
       --check if a beam is a deformGroup trigger, and isn't broken
      if b.deformGroup ~= nil and type(b.deformGroup) ~= "table" and not obj:beamIsBroken(b.cid) then
	    local beamShouldBeIgnored = false
	    for _, g in pairs(brittleGroupsToIgnore) do
	      if string.match(b.deformGroup, g) then --if the group is on the list of groups we should ignore
		    beamShouldBeIgnored = true 
		  end
	    end	   
	    if not beamShouldBeIgnored then --if it matches none of the names to ignore
		  table.insert(beams_with_deformGroups, b) --then add it to the table
	    end
      end
    end
  end
  
  local random_beam = beams_with_deformGroups[math.random(#beams_with_deformGroups)] --select random beam to be broken
  if random_beam == nil then --if there is nothing else to break, don't break anything
    guihooks.message("No brittle pieces left to damage...", 10, "agenty_mechanical_failure_mod", "") --show message
    return 
  end
  
  local random_deformGroup = random_beam.deformGroup --get the selected beam's deformGroup
  beamstate.triggerDeformGroup(random_deformGroup) --trigger deformGroup
  
   --if the beam has a breakGroup, use that as the part name for message, as it's more clear
  if random_beam.breakGroup ~= nil and type(random_beam.breakGroup) ~= "table" then
	guihooks.message(random_beam.breakGroup .. " damage", 10, "agenty_mechanical_failure_mod", "") --show message what broke
  else --otherwise use deformGroup name
    if random_deformGroup == "radiator_damage" then --if radiator got damage this way, it won't leak yet
	  guihooks.message("Radiator damage (no leaking yet)", 10, "agenty_mechanical_failure_mod", "")
	elseif random_deformGroup == "oilpan_damage" then --same with oil pan
	  guihooks.message("Oil pan damage (no leaking yet)", 10, "agenty_mechanical_failure_mod", "")
	else
      guihooks.message(random_deformGroup .. " damage", 10, "agenty_mechanical_failure_mod", "") --show message what broke
	end
  end
  
end


local function deformRandomBeam() --decreases random beam length by random amount

  local deformableBeamsTemp = {} --this will store all beams that can still be deformed
  for _, b in pairs(v.data.beams) do  --cycle through all beams
     --check if a beam isn't broken
    if not obj:beamIsBroken(b.cid) then
	  --we need to make sure that deforming the beam won't break the mod
	  if not (v.data.nodes[b.id1].name == nil and v.data.nodes[b.id1].slidingFrictionCoef == nil) then --check first node
	    if not (v.data.nodes[b.id2].name == nil and v.data.nodes[b.id2].slidingFrictionCoef == nil) then --check second node
	      table.insert(deformableBeamsTemp, b) --then add it to the table
		end
	  end
    end
  end
  
  local deformableBeams = {}
  if engineDeformationOnly == false then
    deformableBeams = deformableBeamsTemp
  else
    for _, b in pairs(deformableBeamsTemp) do  --cycle through all beams
	  for _, g in pairs(genericEngineDamageDeformGroups) do
	    if b.deformGroup ~= nil and type(b.deformGroup) ~= "table" and string.match(b.deformGroup, g) then --if the beam triggers engine damage
	      table.insert(deformableBeams, b) --then add it to the table
		end		
	  end
	end
  end
  
  local deformableBeamsFinal = {}
  if deformRandomBeamIgnoreWheels == false then
    deformableBeamsFinal = deformableBeams
  else
    for _, b in pairs(deformableBeams) do  --cycle through all beams
	  if v.data.nodes[b.id1].name ~= nil and v.data.nodes[b.id2].name ~= nil then --if isn't autogenerated
	    for _, n in pairs(genericWheelCenterNodeNames) do --check if it doesn't attach to wheel center
		  --elaborate workaround to work with the NASCAR mod lol
		  if not v.data.nodes[b.id1].name:match(n) and not v.data.nodes[b.id2].name:match(n) and ((b.breakGroup == nil and type(b.breakGroup) ~= "table") or not (type(b.breakGroup) ~= "table" and (string.match(b.breakGroup, "wheel") or string.match(b.breakGroup, "Wheel") or string.match(b.breakGroup, "axle"))) ) then
	        table.insert(deformableBeamsFinal, b) --then add it to the table
		  end
		end
	  end
	end
  end
  
  local random_beam = deformableBeamsFinal[math.random(#deformableBeamsFinal)] --select random beam to be deformed
  if random_beam == nil then --if there is nothing else to break, don't break anything
    guihooks.message("Nothing left to deform that fits the set options...", 10, "agenty_mechanical_failure_mod", "") --show message
    return 
  end
  
  --emergency check if deforming the beam can break the mod
  if v.data.nodes[random_beam.id1].name == nil and v.data.nodes[random_beam.id1].slidingFrictionCoef == nil then --check first node
    guihooks.message("Error: Beam cannot be deformed safely", 10, "agenty_mechanical_failure_mod", "warning") --show error message
	return
  end
  
  if v.data.nodes[random_beam.id2].name == nil and v.data.nodes[random_beam.id2].slidingFrictionCoef == nil then --check second node
    guihooks.message("Error: Beam cannot be deformed safely", 10, "agenty_mechanical_failure_mod", "warning") --show error message
	return
  end
  
  local deformationRatioPercentage = math.random(100) --percentage of random deformation ratio
  local deformationRatio = 1 - (deformationRatioPercentage / 100) --actual ratio
  obj:setBeamLength(random_beam.cid, obj:getBeamLength(random_beam.cid) * deformationRatio) --set deformed length of selected beam
  
  --we can't use node groups  for beam names as they get removed at load stage for optimiaztion
  local randomBeamPartName = "Random panel" --default name
  if random_beam.beamStrength ~= math.huge then randomBeamPartName = "Random bolt" end --if beam is destructible, we treat it as a bolt
  
  --we will try to find a suitable name for the beam now, if it exists somewhere, otherwise we'll use the above
  if random_beam.partOrigin  ~= nil then randomBeamPartName = "Part of " .. random_beam.partOrigin end --try to get Jbeam part name as default
  
  --if either of beam's nodes has no name/id, then it was generated by the wheel generator, so it's a wheel
  if v.data.nodes[random_beam.id1].name == nil or v.data.nodes[random_beam.id2].name == nil then randomBeamPartName = "Random wheel" end
  if random_beam.deformGroup ~= nil and type(random_beam.deformGroup) ~= "table" then randomBeamPartName = random_beam.deformGroup end --try deform group name first
  if random_beam.breakGroup  ~= nil and type(random_beam.breakGroup) ~= "table"  then randomBeamPartName = random_beam.breakGroup .. " bolt" end --try break group name after
  if random_beam.name        ~= nil then randomBeamPartName = random_beam.name end --try checking if beam has its own name
  if random_beam.soundFile   ~= nil then randomBeamPartName = "Random suspension spring" end --lastly check if beam is a spring
  
  if randomBeamPartName == "radiator_damage" then randomBeamPartName = "Radiator" end --bruteforce fix
  
  guihooks.message(randomBeamPartName .. " deformed by " .. deformationRatioPercentage .. "%", 10, "agenty_mechanical_failure_mod", "") --show message
  
end


local function breakRandomBeam() --breaks random beam

  local breakableBeamsTemp = {} --this will store all beams that can still be broken
  for _, b in pairs(v.data.beams) do  --cycle through all beams
     --check if a beam isn't broken
    if not obj:beamIsBroken(b.cid) then
	  --we need to make sure that breaking the beam won't break the mod
	  if not (v.data.nodes[b.id1].name == nil and v.data.nodes[b.id1].slidingFrictionCoef == nil) then --check first node
	    if not (v.data.nodes[b.id2].name == nil and v.data.nodes[b.id2].slidingFrictionCoef == nil) then --check second node
	      table.insert(breakableBeamsTemp, b) --then add it to the table
		end
	  end
    end
  end
  
  local breakableBeams = {}
  if toggleRealisticDamage == false then
    breakableBeams = breakableBeamsTemp
  else
    for _, b in pairs(breakableBeamsTemp) do  --cycle through all beams
	   --if isn't unbreakable and isn't a dampng or support beam (can't directly check beam type at this stage due to hardcoded optimiaztion stuff)
	  if b.beamStrength ~= math.huge and b.beamSpring ~= 0 and b.beamPrecompression == 1 then
	    table.insert(breakableBeams, b) --then add it to the table
	  end
	end
  end
  
  local breakableBeamsFinal = {}
  if breakRandomBeamIgnoreWheels == false then
    breakableBeamsFinal = breakableBeams
  else
    for _, b in pairs(breakableBeams) do  --cycle through all beams
	  if v.data.nodes[b.id1].name ~= nil and v.data.nodes[b.id2].name ~= nil then --if isn't autogenerated
	    for _, n in pairs(genericWheelCenterNodeNames) do --check if it doesn't attach to wheel center
		  --elaborate workaround to work with the NASCAR mod lol
		  if not v.data.nodes[b.id1].name:match(n) and not v.data.nodes[b.id2].name:match(n) and ((b.breakGroup == nil and type(b.breakGroup) ~= "table") or not (type(b.breakGroup) ~= "table" and (string.match(b.breakGroup, "wheel") or string.match(b.breakGroup, "Wheel") or string.match(b.breakGroup, "axle"))) ) then
	        table.insert(breakableBeamsFinal, b) --then add it to the table
		  end
		end
	  end
	end
  end
  
  local random_beam = breakableBeamsFinal[math.random(#breakableBeamsFinal)] --select random beam to be broken
  if random_beam == nil then --if there is nothing else to break, don't break anything
    guihooks.message("Nothing left to break that fits the current settings...", 10, "agenty_mechanical_failure_mod", "") --show message
    return 
  end
  
  --emergency check if breaking the beam can break the mod
  if v.data.nodes[random_beam.id1].name == nil and v.data.nodes[random_beam.id1].slidingFrictionCoef == nil then --check first node
    guihooks.message("Error: Beam cannot be broken safely", 10, "agenty_mechanical_failure_mod", "warning") --show error message
	return
  end
  
  if v.data.nodes[random_beam.id2].name == nil and v.data.nodes[random_beam.id2].slidingFrictionCoef == nil then --check second node
    guihooks.message("Error: Beam cannot be broken safely", 10, "agenty_mechanical_failure_mod", "warning") --show error message
	return
  end
  obj:breakBeam(random_beam.cid) --break it
  
  --we can't use node groups  for beam names as they get removed at load stage for optimiaztion
  local randomBeamPartName = "Random panel" --default name
  if random_beam.beamStrength ~= math.huge then randomBeamPartName = "Random bolt" end --if beam is destructible, we treat it as a bolt
  
  --we will try to find a suitable name for the beam now, if it exists somewhere, otherwise we'll use the above
  if random_beam.partOrigin  ~= nil then randomBeamPartName = "Part of " .. random_beam.partOrigin end --try to get Jbeam part name as default
  
  --if either of beam's nodes has no name/id, then it was generated by the wheel generator, so it's a wheel
  if v.data.nodes[random_beam.id1].name == nil or v.data.nodes[random_beam.id2].name == nil then randomBeamPartName = "Random wheel" end
  if random_beam.deformGroup ~= nil and type(random_beam.deformGroup) ~= "table" then randomBeamPartName = random_beam.deformGroup end --try deform group name first
  if random_beam.breakGroup  ~= nil and type(random_beam.breakGroup) ~= "table"  then randomBeamPartName = random_beam.breakGroup .. " bolt" end --try break group name after
  if random_beam.name        ~= nil then randomBeamPartName = random_beam.name end --try checking if beam has its own name
  if random_beam.soundFile   ~= nil then randomBeamPartName = "Random suspension spring" end --lastly check if beam is a spring
  
  if randomBeamPartName == "radiator_damage" then randomBeamPartName = "Radiator" end --bruteforce fix
  
  guihooks.message(randomBeamPartName .. " damaged", 10, "agenty_mechanical_failure_mod", "") --show message
  
end


local function deflateRandomTire() --punctures a random tire (breaks a random beam in it)
	
  if lowerTireDeflationChance == true then
    local chanceOfDeflation = math.random(2) --50% chance
	if chanceOfDeflation == 1 then
      guihooks.message("You are lucky, nothing happened", 10, "agenty_mechanical_failure_mod", "tag_faces") --show message
      return
	end
  end
	
  if canDeflateTires == false then --if deflating tires will break the game, then don't do it
    local warning = "This function is not supported for this vehicle configuration. Try using different wheels/tires if possible."
    guihooks.message(warning , 10, "agenty_mechanical_failure_mod", "warning") --show warning message
    return
  end
  
  if hasNoTires == true then --either uses an outdated tire system, or has no tires, or most likely both. Or could be a fake tire like the tire stack prop
    guihooks.message("Can't puncture tires" , 10, "agenty_mechanical_failure_mod", "")
  end --if there are tires, they will be deflated later, overriding the message
  
  if wheels.wheels == nil then 
    guihooks.message("Error: No tires detected" , 10, "agenty_mechanical_failure_mod", "warning") --show warning message
    return 
  end --check if the vehicle even has any wheels
  
  local random_tire = wheels.wheels[math.random(#wheels.wheels + 1) - 1] --select random wheel number, counted from 0
  
  for i, w in pairs(wheels.wheels) do --for each wheel
    if w == random_tire then  --if random number matches the wheel number
	  local wheel = v.data.wheels[i] --get wheel on Jbeam side
	  local tireBeams = {} --this will store all beams of the tire that aren't broken yet
	  --for different types of beams in the tire: check if exist, check if not broken, add to array
	  if wheel.treadBeams ~= nil then  --tread beams
	    for _, b in pairs(wheel.treadBeams) do
		  if not obj:beamIsBroken(b) then
		    table.insert(tireBeams, b)
		  end
		end
	  end
	  if wheel.sideBeams ~= nil then  --side beams
	    for _, b in pairs(wheel.sideBeams) do
		  if not obj:beamIsBroken(b) then
		    table.insert(tireBeams, b)
		  end
		end
	  end
	  if wheel.peripheryBeams ~= nil then  --periphery beams
	    for _, b in pairs(wheel.peripheryBeams) do
		  if not obj:beamIsBroken(b) then
		    table.insert(tireBeams, b)
		  end
		end
	  end
	  if wheel.reinfBeams ~= nil then  --reinf beams
	    for _, b in pairs(wheel.reinfBeams) do
		  if not obj:beamIsBroken(b) then
		    table.insert(tireBeams, b)
		  end
		end
	  end
	  if wheel.pressuredBeams ~= nil then  --pressured beams
	    for _, b in pairs(wheel.pressuredBeams) do
		  if not obj:beamIsBroken(b) then
		    table.insert(tireBeams, b)
		  end
		end
	  end
	  local random_beam = tireBeams[math.random(#tireBeams)] --select random beam to be broken
	  if random_beam == nil then --the tire either is fully shredded, or outdated, or is a square tire or other weird tire, or there is no tire at all...
		guihooks.message(wheel.name .. " tire can't be punctured...", 10, "agenty_mechanical_failure_mod", "") --show message
		return 
	  end
	  obj:breakBeam(random_beam) --puncture tire
	  guihooks.message(wheel.name .. " tire punctured", 10, "agenty_mechanical_failure_mod", "") --show message what broke
	end
  end
	
end


local function igniteRandomPart() --ignites a random node, with a chance of exploding the car

  if canDeflateTires == false then --fire can deflate tires
    local warning = "This function is not supported for this vehicle configuration. Sorry for inconvenience."
    guihooks.message(warning , 10, "agenty_mechanical_failure_mod", "warning") --show warning message
    return
  end
  
  local wheelNodes = {} --will store wheel nodes as they are an exception from flammable
  if wheels.wheels then --if wheels exist
    for id, wd in pairs(wheels.wheels) do --add all wheel nodes to array
      wheelNodes[wd.node1] = {wheelID = id, node1 = wd.node1, node2 = wd.node2}
      wheelNodes[wd.node2] = {wheelID = id, node1 = wd.node1, node2 = wd.node2}
    end
  end
  
  local flammableNodes = {} --will store all flammable nodes
  for k, n in pairs(v.data.nodes) do --for each node
    if n.flashPoint ~= nil and not wheelNodes[k] then --node is flammable and is not part of wheel
      table.insert(flammableNodes, k) --add to array
    end
  end
  
  if #flammableNodes <= 0 then --if nothing is flammmable
    guihooks.message("Nothing left to ignite..." , 10, "agenty_mechanical_failure_mod", "warning") --show message
    return
  end
  
  fire.igniteRandomNode() --this function from vanilla game comes in very handy
  
  local chanceOfExploding
  if toggleRandomExplosion == true then
    chanceOfExploding = math.random(10) --10% chance of car exploding
  else
    chanceOfExploding = 0
  end
  
  if chanceOfExploding == 1 then
	fire.explodeVehicle() --another handy function
	beamstate.breakAllBreakgroups() --break everything in the car
	for _, b in pairs(v.data.beams) do  --cycle through all beams
       --check if a beam is a deformGroup trigger
      if b.deformGroup ~= nil and type(b.deformGroup) ~= "table" then
	    beamstate.triggerDeformGroup(b.deformGroup) --trigger every deformGroup
      end
    end
	guihooks.message("The car has randomly exploded!", 10, "agenty_mechanical_failure_mod", "warning") --show fancy message
  else
	guihooks.message("Random part caught on fire!", 10, "agenty_mechanical_failure_mod", "warning") --show fancy message
  end

end


local function toggleDamageTimer(seconds) --toggles randomizing damage every set time of seconds

  timeBetweenFailures = seconds or 10 --set amount of seconds
  if type(timeBetweenFailures) == "table" then timeBetweenFailures = 10 end --in case someone somehow found a way to manually input the value and broke it

  if damageTimerEnabled == true then 
	damageTimerEnabled = false
	guihooks.message("Random damage timer mode disabled", 10, "agenty_mechanical_failure_mod", "")
  else
    damageTimerEnabled = true
	guihooks.message("Random damage timer mode enabled every " .. timeBetweenFailures .. " seconds", timeBetweenFailures, "agenty_mechanical_failure_mod", "")
  end
	
end


--------------------------------------------
----- FUNCTION RANDOMIZER BY ANGELO234 -----
--------------------------------------------


local fullFunctionsLists = {
    ['breakRandomPart'] = breakRandomPart,
	['deformRandomPart'] = deformRandomPart,
	['igniteRandomPart'] = igniteRandomPart,
	['breakRandomBeam'] = breakRandomBeam,
	['deformRandomBeam'] = deformRandomBeam,
	['deflateRandomTire'] = deflateRandomTire,
}

local functions = {}

local function addFunction(func)

    if fullFunctionsLists[func] then
        functions[func] = fullFunctionsLists[func]
    end
	
end

local function removeFunction(func)

    functions[func] = nil
	
end

local function getFunctions()

    return functions
	
end

local function callFunction(func)

    if functions[func] then
        functions[func]()
    end
	
end

local function callRandomFunction()

    functions[tableKeys(functions)[math.random(tableSize(functions))]]()
	
end

local function onExtensionLoaded()

end


--------------------------------------------


local function mechanicalFailure() --randomizes the mechanical failure

    if tableSize(functions) == 0 then return end	
	callRandomFunction()
	
end


local function onInit() --this basically checks for everything that could possibly go wrong

  damageTimerEnabled = false
  damageTimer = 0
  canDeflateTires = true
  hasNoTires = false
  timeBetweenFailures = 10
  local warning --prevents overriding warnings
  
  --we need to perform some checks to see if deflating tires won't break the game
  if wheels.wheels == nil or v.data.wheels == nil then  --if wheels are not initiated for some reason
    local warning = "Unknown error with wheels/tires. Some functions of the Mechanical Failure Mod might not work properly."
    guihooks.message(warning, 10, "agenty_mechanical_failure_mod", "warning") --show warning message
    canDeflateTires = false 
  else --wheels initiated...
    if #wheels.wheels == 0 or #v.data.wheels == 0 then  --...but not used in the vehicle
	  canDeflateTires = true --still allow to ignite the vehicle
	  hasNoTires = true --but show warning anyway
	  warning = "No wheels detected for this vehicle/prop. Some functions of the Mechanical Failure Mod might not work properly."
      guihooks.message(warning, 10, "agenty_mechanical_failure_mod", "warning") --show warning message
	  return --don't check further, otherwise the game will think something else broke
	else --there are wheels but we aren't sure if there are tires
	  for _, wheel in pairs(v.data.wheels) do --for each wheel
	    if not wheel.pressuredBeams then --most likely has no tire, OR it is a square wheel, OR it's Pigeon's stabilizer wheel or some other weird stuff
		  canDeflateTires = true --for a rare case where it has the square wheels equipped, they can't deflate but otherwise act like normal wheels
		  hasNoTires = true --this is only to show the warning then
		end
	  end
	end
  end --at this point we are sure the vehicle has wheels. But they could have no tires, could be using a weird system, or could be broken
  
  for _, wheel in pairs(v.data.wheels) do --for each wheel
	for _, nodecid in pairs(wheel.nodes) do --then for each wheel node
	  if v.data.nodes[nodecid].slidingFrictionCoef == nil then --check if the tire mod is broken
        obj:setNodeFrictionSlidingCoefs(nodecid, 1, 1) --attempt to fix it, won't work in most cases
	  end
	  if v.data.nodes[nodecid].slidingFrictionCoef == nil then --check if still broken. At this point it might as well be tank treads instead of tire
		canDeflateTires = false --mod is broken, we can't deflate the tires
		--warning moved here, so it doesn't get overwritten by other warnings...
		warning = "WARNING! The mod you are using for either this vehicle, or its wheels or tires, is either broken/outdated, or uses a really unusual wheel system, or is missing some tires. Some functions of the Mechanical Failure Mod are not supported for this configuration. Try replacing the wheels/tires or use a different vehicle, otherwise you might experience issues!"
		guihooks.message(warning, 10, "agenty_mechanical_failure_mod", "warning") --show warning message
	  end
	end
  end

end


local function updateGFX(dt) --triggers random damage every set amount of seconds

  if damageTimerEnabled == false then return end --do nothing if the thing is not enabled in the first place
  damageTimer =  damageTimer + dt --count every second
  if damageTimer >= timeBetweenFailures then --time passed
    mechanicalFailure() --trigger random mechanical failure
	damageTimer = 0 --reset timer
  end

end


-- public interface

M.addFunction           = addFunction
M.removeFunction        = removeFunction
M.getFunctions          = getFunctions
M.callFunction          = callFunction
M.callRandomFunction    = callRandomFunction
M.onExtensionLoaded     = onExtensionLoaded

M.setSettings           = setSettings
M.unsetSettings         = unsetSettings
M.setWheelSettings      = setWheelSettings
M.unsetWheelSettings    = unsetWheelSettings
M.mechanicalFailure     = mechanicalFailure
M.breakRandomPart       = breakRandomPart
M.deformRandomPart      = deformRandomPart
M.deformRandomBeam      = deformRandomBeam
M.breakRandomBeam       = breakRandomBeam
M.breakRandomBeamSafe   = breakRandomBeamSafe
M.deflateRandomTire     = deflateRandomTire
M.igniteRandomPart      = igniteRandomPart
M.toggleDamageTimer     = toggleDamageTimer
M.updateGFX             = updateGFX
M.onReset               = onInit
M.onInit                = onInit

return M
